/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "globals.h"
#include "acache.h"
#include "article.h"
#include "article-db.h"
#include "debug.h"
#include "log.h"
#include "message-window.h"
#include "nntp.h"
#include "queue.h"
#include "queue-item-body.h"
#include "text.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static gint queue_item_body_run (Task* item);

static gchar* queue_item_body_describe (const StatusItem* item);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

static void
queue_item_body_destructor (PanObject *o)
{
	QueueItemBody *item = QUEUE_ITEM_BODY(o);

	if (item->checkout) {
		item->checkout = FALSE;
		acache_file_checkin(item->message->message_id);
	}

	if (item->message) {
		message_free (item->message);
		item->message = NULL;
	}

	queue_item_destructor (o);
}

/************
*************  PUBLIC ROUTINES
************/

PanObject*
queue_item_body_new (server_data* sdata, group_data* gdata, Message* message)
{
	QueueItemBody* body = NULL;
	gboolean needs_socket;

	/* sanity checks */
	g_return_val_if_fail (message!=NULL, NULL);
	g_return_val_if_fail (message->message_id!=NULL, NULL);
	g_return_val_if_fail (*message->message_id!='\0', NULL);

	/* do we need to pull this body from the network? */
	needs_socket = TRUE;
	if (acache_body_exists(message->message_id) &&
	    acache_header_exists(message->message_id))
	{
        	debug (DEBUG_PAN_OBJECT, "queue_item_body_new: no socket needed; it's cached");
		acache_file_checkout(message->message_id);
		needs_socket = FALSE;
	}

	/* create the queue item */
       	body = g_new0 (QueueItemBody, 1);
        debug (DEBUG_PAN_OBJECT, "queue_item_body_new: %p", body);
	queue_item_constructor (QUEUE_ITEM(body),
		queue_item_body_destructor,
		queue_item_body_describe,
		queue_item_body_run,
		sdata, TRUE, needs_socket);
	TASK(body)->is_article_reader = TRUE;


	/* initialize the queue-item-body */
	body->gdata = gdata;
	body->message = message;
	body->checkout = !needs_socket;

	return PAN_OBJECT(body);
}

/*****
******
*****/

static gchar*
queue_item_body_describe (const StatusItem* si)
{
	QueueItemBody *item = QUEUE_ITEM_BODY(si);

	if (item->checkout)
	{
		return g_strdup_printf (_("Loading body of cached article ``%s'' from group %s"),
			item->message->message_id,
			item->gdata->name );
	} else {
		return g_strdup_printf (_("Downloading body of ``%s'' from group %s"),
			item->message->message_id,
			item->gdata->name );
	}
}

/*****
******
*****/

static gint
queue_item_body_run (Task* item)
{
	server_data *sdata = item->sdata;
	group_data *gdata = QUEUE_ITEM_BODY(item)->gdata;
	Message *msg = QUEUE_ITEM_BODY(item)->message;
	article_db db = NULL;
	article_data *adata = NULL;

	/* sanity checks */
	g_return_val_if_fail(sdata!=NULL, -1);
	g_return_val_if_fail(gdata!=NULL, -1);
	g_return_val_if_fail(msg!=NULL, -1);
	g_return_val_if_fail(msg->message_id!=NULL, -1);
	g_return_val_if_fail(*msg->message_id!='\0', -1);

	if (QUEUE_ITEM_BODY(item)->checkout)
		status_item_emit_status(STATUS_ITEM(item),
			_("Loading Cached Article '%s'"),
			msg->message_id );

	/* get the article header */
	db = ahdb_ref (sdata, gdata);
	adata = ahdb_get ( db, (const char*)msg->message_id );
	ahdb_unref ( db );

	/* make sure we've got the first header */
	if ( !adata ) {
		status_item_emit_error(STATUS_ITEM(item),
			_("Couldn't find article header for message_id '%s'"),
			msg->message_id );
		return -1;
	}

	article_remove_flag (sdata, gdata, adata, STATE_DOWNLOAD_QUEUED, TRUE);

	/* Check the cache before we download */
	if (!acache_header_exists(adata->message_id) ||
            !acache_body_exists(adata->message_id))
	{
		acache_file_checkout (adata->message_id);
		if (nntp_article_download(
			STATUS_ITEM(item), QUEUE_ITEM(item)->sock,
			sdata, gdata, adata, &QUEUE_ITEM(item)->hint_abort))
		{
			status_item_emit_status(STATUS_ITEM(item),
				_("Downloading article ``%s'' failed."),
				adata->subject);
			acache_file_checkin (adata->message_id);
			article_free (adata);

			/* if user's expecting a message, let them know it's off. */
			if (item->high_priority && STATUS_ITEM(item)->last_error!=NULL)
				update_body_pane (Pan.text, STATUS_ITEM(item)->last_error);

			return -1;
		}
	}

	/* Display, et cetera */
	if (item->high_priority)
	{
		message_import_adata (msg, adata);
		article_mark_read (sdata, gdata, adata, TRUE);
		text_update (msg);

		/* check in this file.. */
		if (QUEUE_ITEM_BODY(item)->checkout) {
			QUEUE_ITEM_BODY(item)->checkout = FALSE;
			acache_file_checkin (adata->message_id);
		}

	       	/* text_update frees the message */
		QUEUE_ITEM_BODY(item)->message = msg = NULL;
	}

	/* cleanup */
	acache_file_checkin (adata->message_id);
	article_free (adata);
	adata = NULL;
	return 0;
}
