/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "article.h"
#include "article-db.h"
#include "date.h"
#include "debug.h"
#include "log.h"
#include "queue-item-headers.h"

#include "articlelist.h"
#include "grouplist.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static gint queue_item_headers_run (QueueItem* item);

static gchar* queue_item_headers_describe (const StatusItem *item);

static int nntp_articlelist_download (QueueItemHeaders* item, GPtrArray* setme);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

/*****
******
*****/

PanObject*
queue_item_headers_new (server_data *sdata,
	                gboolean high_priority,
			group_data *gdata,
			HeaderDownloadType dl_type)
{
	QueueItemHeaders *item = g_new0 (QueueItemHeaders, 1);
	gboolean need_socket = dl_type != HEADERS_LOCAL;

	debug (DEBUG_PAN_OBJECT, "queue_item_headers_new: %p", item);

//	if ((gdata->type == USER_FOLDER) && (need_socket))
//		return NULL;

	queue_item_constructor (QUEUE_ITEM(item),
		queue_item_destructor,
		queue_item_headers_describe,
		queue_item_headers_run, sdata, high_priority, need_socket);

	item->gdata = gdata;
	item->download_type = dl_type;

	return PAN_OBJECT(item);
}


/*****
******
*****/

static int
pp_adata_message_id_comp (const void* va, const void* vb )
{
	const article_data* a = *(const article_data**)va;
	const article_data* b = *(const article_data**)vb;
	return strcmp ( a->message_id, b->message_id );
}

static gint
queue_item_headers_run (QueueItem* item)
{
	QueueItemHeaders *item_h = QUEUE_ITEM_HEADERS(item);
	GPtrArray * articles = NULL;
	GPtrArray * old_articles = NULL;
	article_db db = NULL;
	group_data *gdata = NULL;
	server_data *sdata = NULL;
	gboolean active;
	gboolean free_segment = TRUE;
	int read;
	int i;
	int worst;

	if (item_h->download_type == HEADERS_LOCAL) {
		articlelist_load (item->sdata, item_h->gdata, STATUS_ITEM(item));
		return 0;
	}

	sdata = item->sdata;
	gdata = item_h->gdata;

	/* get the article headers... */
	articles = g_ptr_array_new ();
	i = nntp_articlelist_download (QUEUE_ITEM_HEADERS(item), articles);
	if (i)
	{
		int foo = 0;
		for (foo=0; foo<articles->len; ++foo)
			article_free (ARTICLE_DATA(g_ptr_array_index(articles,foo)));
		g_ptr_array_free (articles, TRUE);
		return i;
	}

	/* no articles! */
	if (!articles->len)
	{
		status_item_emit_status (STATUS_ITEM(item),
			_("No articles retrieved for group %s"),
			gdata->name);

		g_ptr_array_free (articles, TRUE);
		return 0;
	}

	log_add_va (
		_("Merging %d article headers with old headers from %s"),
		articles->len,
		gdata->name);

	active = sdata==articlelist_get_current_server() &&
	         gdata==articlelist_get_current_group();


	/* merge with existing articles (1) calculate # of steps */
	db = ahdb_ref (sdata, gdata);
	i = ahdb_length(db); /* FIXME: very wasteful */
	worst = articles->len;
	if (QUEUE_ITEM_HEADERS(item)->download_type != HEADERS_ALL)
		worst += i;
	status_item_emit_init_steps (STATUS_ITEM(item),
		i                         /* load old headers */
		+ articles->len           /* merge states */
		+ worst                   /* worst-case on save */
		+ (active?(worst*3):0));  /* worse-case on alist_set_contents */

	/* merge with existing articles (2) load old articles. */
	status_item_emit_status (STATUS_ITEM(item), _("Loading Old Headers"));
	old_articles = ahdb_get_all (db, status_item_next_step_gfunc, item);
	qsort (old_articles->pdata, old_articles->len, sizeof(gpointer*), pp_adata_message_id_comp);

	/* merge with existing articles (3) merge states.
	   This is so that read messages don't suddenly become unread,
	   and so forth. */
	status_item_emit_status (STATUS_ITEM(item), _("Merging Headers"));
	for (i=0; i!=articles->len; ++i)
	{
		gboolean exact = FALSE;
		article_data* adata = ARTICLE_DATA(g_ptr_array_index(articles,i));

		const int index = lower_bound (&adata,
			old_articles->pdata, old_articles->len,
			sizeof(article_data*),
			pp_adata_message_id_comp,
			&exact);

		if (exact) { /* got a match! */
			article_data* a = ARTICLE_DATA(g_ptr_array_index(old_articles,index));
			adata->state = a->state;
			article_free (a);
			g_ptr_array_remove_index (old_articles, index);
		}
		status_item_emit_next_step (STATUS_ITEM(item));
	}

	/* merge with existing articles (4) get list of all existing articles */
	for (i=0; i<old_articles->len; ++i)
	{
		article_data* a = ARTICLE_DATA(g_ptr_array_index(old_articles,i));
		if (QUEUE_ITEM_HEADERS(item)->download_type == HEADERS_ALL) {
			article_free (a);
		} else {
			g_ptr_array_add (articles, a);
		}
	}
	g_ptr_array_free (old_articles, TRUE);
	old_articles = NULL;

	/* merge with existing articles (5) update the database */
	read = 0;
	for (i=0; i!=articles->len; ++i) {
		if (article_flag_on (ARTICLE_DATA(g_ptr_array_index(articles,i)), STATE_READ))
			++read;
	}
	ahdb_erase_all (db);
	group_set_attrib_i (sdata, gdata, "Read", read);
	group_set_attrib_i (sdata, gdata, "Total", articles->len);
	status_item_emit_status (STATUS_ITEM(item), _("Saving Headers"));
	ahdb_save_all (db, (article_data**)articles->pdata, articles->len, status_item_next_step_gfunc, item);

	/* merge with existing articles (5) update UI */
	grouplist_update_row (gdata);
	if (active)
	{
		articlelist_set_contents (
			sdata, gdata,
			(article_data**)articles->pdata, articles->len,
			STATUS_ITEM(item));
		free_segment = FALSE; /* cause articlelist owns it now */
	}

	/* cleanup */
	g_ptr_array_free (articles, free_segment);
	status_item_emit_progress (STATUS_ITEM(item), 0);
	ahdb_sync (db);
	ahdb_unref (db);

	return 0;
}

/*****
******
*****/

static int
nntp_articlelist_download (QueueItemHeaders* item, GPtrArray* setme)
{
	const gboolean * const abort = &QUEUE_ITEM(item)->hint_abort;
	PanSocket *sock = QUEUE_ITEM(item)->sock;
	server_data *sdata = QUEUE_ITEM(item)->sdata;
	group_data *gdata = QUEUE_ITEM_HEADERS(item)->gdata;
	HeaderDownloadType dl_type = QUEUE_ITEM_HEADERS(item)->download_type;

	const char *buffer = NULL;
	article_data *adata = NULL;
	gint total, total_new;
	int steps=0;
	int first=0, last=0, p_last=0;
	gboolean read_status = 0;

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, -1);
	g_return_val_if_fail (sdata!=NULL, -1);
	g_return_val_if_fail (gdata!=NULL, -1);

	/* change to the right group */
	if (*abort || pan_socket_putline_va (sock, "GROUP %s\r\n", gdata->name)) {
		queue_item_emit_sockwrite_err (QUEUE_ITEM(item));
		return -1;
	}
	if (*abort || pan_socket_getline (sock, &buffer)) {
		queue_item_emit_sockread_err (QUEUE_ITEM(item));
		return -1;
	}
	if (strtol(buffer, NULL, 10) != 211) {
		status_item_emit_error (STATUS_ITEM(item),
			_("(Couldn't select group '%s'"), gdata->name);
		return -1;
	}
	sscanf (buffer, "%*d %d %d %d", &total, &first, &last);

	p_last = group_get_attrib_i (sdata, gdata, "Last");

	total_new = last - p_last;

	if (total == '0') {
		status_item_emit_status (STATUS_ITEM(item),
			_("No articles found for group '%s'"), gdata->name);
		return -1;
	}

	if ((dl_type == HEADERS_NEW) && p_last)
	{
		if (p_last == last)
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("No new articles found for group '%s'"),
				gdata->name);
			return 0;
		}
		else
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting new article headers..."));

			if (*abort || pan_socket_putline_va (sock,
					"XOVER %d-%d\r\n", p_last, last))
			{
				queue_item_emit_sockwrite_err(QUEUE_ITEM(item));
				return -1;
			}
		}
	}
	else
	{
		if ((total > 150) && ((dl_type==HEADERS_SAMPLE)))
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting sample"));
			last = first + 150;
		}
		else
		{
			status_item_emit_status (STATUS_ITEM(item),
				_("Getting all article headers"));
		}

		if (*abort || pan_socket_putline_va (sock,"XOVER %d-%d\r\n",first,last))
		{
			queue_item_emit_sockwrite_err (QUEUE_ITEM(item));
			return -1;
		}
	}

	status_item_emit_status (STATUS_ITEM(item),
		_("Waiting for Response from Server..."));

	if (*abort || pan_socket_getline (sock, &buffer)) {
		queue_item_emit_sockread_err (QUEUE_ITEM(item));
		return -1;
	}

	if (strtol(buffer, (char **) NULL, 10) != 224) {
		status_item_emit_error (STATUS_ITEM(item),
			_("Couldn't retrieve headers\nServer said: %s"), buffer);
		return -1;
	}

	steps = ((dl_type==HEADERS_NEW) && (group_get_attrib_i (sdata, gdata,"Total")>0))
		? total_new
		: total;
	status_item_emit_init_steps (STATUS_ITEM(item), steps);

	while (!*abort && !((read_status = pan_socket_getline (sock, &buffer))))
	{
		int i = 0;
		int line_qty = 0;
		int part = 0;
	        int parts = 0;
		gchar **sd = NULL;
		gchar *s = NULL;

		/* handle end-of-list */
		if (!strncmp(buffer, ".\r\n", 3))
			break;

		/* split the header */
		sd = g_strsplit (buffer, "\t", 8);

		/* check for incomplete header */
		line_qty = 0;
		while ( sd[line_qty]!=NULL )
			++line_qty;

		/* make sure we have enough headers, and do a minimum
		   sanity check on the headers we did get. */
		if (line_qty<8 || !*sd[0] || !*sd[1] || !*sd[2] || !sd[3] || !*sd[4]) {
			g_message ( "Corrupt Header");
			status_item_emit_error (STATUS_ITEM(item),
				_("Corrupt header ``%s'' skipped"),buffer);
			g_strfreev (sd);
			continue;
		}

		/* create the article data */
		adata = article_new ();
		adata->number = g_strdup(sd[0]);
		adata->subject = g_strdup(sd[1]);
		adata->author = g_strdup(sd[2]);
		adata->date = parse_date (sd[3]);
		adata->message_id = g_strdup(sd[4]);
		adata->references = g_strdup(sd[5]);
		adata->linecount = strtol(sd[7], NULL, 10);
		adata->state = STATE_NEW;

		/* cleanup */
		g_strfreev (sd);

		/* Look for the (n/N) or [n/N] construct in subject lines,
		 * starting at the end of the string and working backwards */
		part = 0;
		parts = 0;
		s = adata->subject;
		for (i = strlen(s) - 1; i > 0; --i) {
			if ((s[i] == ')') || (s[i] == ']')) {
				for (--i; isdigit((int)s[i]); --i);
				if (s[i] == '/') {
					parts = atoi(s + i + 1);
					for (--i; isdigit((int)s[i]); --i);
					if ((s[i] == '(') || (s[i] == '[')) {
						part = atoi(s + i + 1);
						break;
					}
				}
			}
		}

		/* if not a multipart yet, AND if it's a big message, AND
		   it's either in one of the pictures/fan/sex groups or it
		   has commonly-used image names in the subject, guess it's
		   a single-part binary */
		if (!parts
			&& adata->linecount>400
			&& (((strstr(gdata->name,"pictures")
					|| strstr(gdata->name,"fan")
				        || strstr(gdata->name,"mag")
					|| strstr(gdata->name,"sex")))
				|| ((strstr(s,".jpg") || strstr(s,".JPG")
					|| strstr(s,".jpeg") || strstr(s,".JPEG")
					|| strstr(s,".gif") || strstr(s,".GIF")
					|| strstr(s,".tiff") || strstr(s,".TIFF")))))
			part = parts = 1;

		/* Verify Multipart info */
		if ((parts>=1) && (part<=parts)) {
			/* We found one! */
			adata->parts = parts;
			adata->part = part;
		}
		else {
			adata->parts = 0;
			adata->part = 0;
		}

		/* Add the article to the article list */
		g_ptr_array_add (setme, adata);

		/* update the count & progress ui */
		status_item_emit_next_step (STATUS_ITEM(item));
		if (!(setme->len % 29)) {
			status_item_emit_status (STATUS_ITEM(item),
				((dl_type==HEADERS_NEW)
				 ? _("Getting new article headers %d of %d")
				 : _("Getting article headers %d of %d")),
				setme->len, steps);
		}
	}
	if (read_status!=0 || QUEUE_ITEM(item)->hint_abort)
	{
		if (read_status!=0)
			queue_item_emit_sockread_err (QUEUE_ITEM(item));
		/*FIXME: what on timeout, some is better than nothing -- but should set number
		  so that "get new" will pick up from this point.
		  g_slist_foreach (alist, article_free_gfunc, NULL);
		  g_slist_free (alist);
		*/
		return -1;
	}
	
	group_set_attrib_i (sdata, gdata, "First", first);
	group_set_attrib_i (sdata, gdata, "Last", last);

	return 0;
}

static gchar*
queue_item_headers_describe (const StatusItem* item)
{
	const gchar* action = NULL;

	/* sanity checks */
	g_return_val_if_fail (item!=NULL, NULL);

	switch (QUEUE_ITEM_HEADERS(item)->download_type)
	{
		case HEADERS_LOCAL: action=_("cached headers"); break;
		case HEADERS_ALL: action=_("all headers"); break;
		case HEADERS_NEW: action=_("new headers"); break;
		case HEADERS_SAMPLE: action=_("a sample of headers"); break;
		default: pan_warn_if_reached(); action=_("BUG IN CODE"); break;
	}

	return g_strdup_printf ( _("Getting %s for `%s'"),
		action,
		QUEUE_ITEM_HEADERS(item)->gdata->name );
}
